/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <errno.h>

#ifndef O_SHLOCK
#ifdef O_NSHARE
#define O_SHLOCK O_NSHARE
#else
#define O_SHLOCK 0
#endif
#endif

extern char *crypt(const char *, const char *);

struct udb_data {                       /* Data being held in external db */
        time_t          time;           /* allocation time                */
        u_long          addr;           /* assigned address               */
};

static  char    dbname[FILENAME_MAX];
static  DB      *db;
int     verbose = 0;
int	tacspec = 0;

void    edit(char *), dbappend(), opendb(int),
	change_pass(int, char *, char *), error();

int main(ac, av)
	int     ac;
	char   *av[];
{
	int     c;
	DBT     k[1], d[1];
	char    *p;
	int	i;
	struct in_addr iptmp;

	if (ac < 3)
		goto usage;
	av++;   ac--;

	if ((p = (char *) strrchr(*av, '.')) && !strcmp(p, ".db"))
		(void) strcpy(dbname, *av);
	else
		(void) sprintf(dbname, "%s.db", *av);

	while ((c = getopt(ac, av, "ve:p:P:LCad:g:i")) != -1) {
		switch (c) {

		    case 'v':
			verbose++;
			break;

		    case 'i':
			tacspec++;
			break;

		    case 'e':                   /* edit entry */
			opendb(O_RDWR);
			if (optind != ac)
				goto usage;
			edit(optarg);
			exit(0);

		    case 'P':
		    case 'p':                   /* change password */
			opendb(O_RDWR);
			if (optind == ac)
				change_pass(c == 'P', optarg, NULL);
			else if (optind == (ac - 1))
				change_pass(c == 'P', optarg, av[optind]);
			else
				goto usage;
			exit(0);

		    case 'L':                   /* database dump */
			opendb(O_RDONLY);
			while(!(*db->seq) (db, k, d, R_NEXT)) {
				(void) fwrite(k->data, 1, k->size, stdout);
				if (tacspec) {
					fputc(' ',stdout);
					for (i=k->size; i<10; i++) fputc(' ',stdout);
					iptmp.s_addr=((struct udb_data *) d->data)->addr;
					fprintf(stdout, "time=%i\taddr=%s\n", (int)((struct udb_data *) d->data)->time, inet_ntoa(iptmp));
				} else {
					(void) putc('\t', stdout);
					(void) fwrite(d->data, 1, d->size, stdout);
					(void) putc('\n', stdout);
				}
			}
			(*db->close) (db);
			exit(0);

		    case 'C':                   /* create db    */
			if (optind != ac)
				goto usage;
			if (!access(dbname, F_OK) && unlink(dbname) < 0)
				error(1, "can't unlink old db %s\n", dbname);
			if (verbose)
				error(0, "creating db %s\n", dbname);
			opendb(O_RDWR|O_CREAT);
			dbappend();
			exit(0);

		    case 'a':                   /* append to db */
			if (optind != ac)
				goto usage;
			opendb(O_RDWR);
			dbappend();
			exit(0);

		    case 'd':                   /* delete key  */
			optind--;
			opendb(O_RDWR);
			do {
				k->size = strlen(av[optind]);
				k->data = av[optind];
				if (verbose)
					error(0, "deleting key %s\n", k->data);
				if ((*db->del) (db, k, NULL))
					error(0, "unable to delete key %s\n", k->data);
				optind++;
			} while (optind < ac);
			(*db->close) (db);
			exit(0);

		    case 'g':                   /* get key     */
			optind--;
			opendb(O_RDONLY);
			do {
				k->size = strlen(av[optind]);
				k->data = av[optind];
				if ((*db->get) (db, k, d, NULL))
					error(0, "can't get key %s\n", k->data);
				else {
					(void) fwrite(k->data, 1, k->size, stdout);
					if (tacspec) {
						fputc(' ',stdout);
						for (i=k->size; i<10; i++) fputc(' ',stdout);
						iptmp.s_addr=((struct udb_data *) d->data)->addr;
						fprintf(stdout, "time=%i\taddr=%s\n", (int)((struct udb_data *) d->data)->time, inet_ntoa(iptmp));
					} else {
						(void) putc('\t', stdout);
						(void) fwrite(d->data, 1, d->size, stdout);
						(void) putc('\n', stdout);
					}
				}
				optind++;
			} while (optind < ac);
			(*db->close) (db);
			exit(0);

		     default:
usage:
			error(1, "Usage: tdb dbname [OPTION]\n"
			"  -a append to db\n"
			"  -C create db\n"
			"  -d key ... delete key\n"
			"  -e key edit entry\n"
			"  -g key ... get key\n"
			"  -i specific for dynipd db\n"
			"  -L database dump\n"
			"  -p|-P user [passw] change password\n"
			"  -v verbose\n");
		}
	}
	goto usage;
}

/*
 * Open database
 */
void    opendb(mode)
	int     mode;
{
	db = dbopen(dbname, mode|O_SHLOCK, 0600, DB_HASH, NULL);

	if (!db)
		error(1, "can't %s database %s\n", (mode & O_CREAT)?
			"create":"open", dbname);
}

/*
 * Append to database
 */
void    dbappend()
{
	unsigned char   iline[BUFSIZ], *p;
	char *p1;
	DBT             k[1], d[1];
	int             line = 0;
	struct udb_data element;

	if (verbose)
		error(0, "append to db %s\n", dbname);

	while(fgets(iline, BUFSIZ-1, stdin)) {
		line++;
		if (*iline == '#')
			continue;       /* comment */
		iline[BUFSIZ]='\0';
		for (p = iline; *p && *p != '\n'; p++)
			;
		*p = '\0';
		for (p = iline; *p != '\t' && *p != ' ' && *p; p++)
			;
		if (!*p) {
badline:
			error(0, "bad input line %d; ignored\n", line);
			continue;
		}
		if (tacspec) {
			*p = '\0';
			for(p++; *p && (*p == ' ' || *p == '\t'); p++);
			if (!*p) goto badline;
			if (strncmp(p,"time=",5)) goto badline; p+=5;
			errno=0;
			element.time=strtoul(p,&p1,10);
			if (errno != 0) goto badline;
			p=p1;
			for(p++; *p && (*p == ' ' || *p == '\t'); p++);
			if (!*p) goto badline;
			if (strncmp(p,"addr=",5)) goto badline; p+=5;
			element.addr=inet_addr(p);
			if (element.addr == -1) goto badline;
			if (verbose)
				error(0, "adding key=%s time=%i, addr=0x%08X\n", iline, element.time, element.addr);
		} else {
			*p++ = '\0';
			while(*p == ' ' || *p == '\t')
				p++;
			if (!*p)
				goto badline;
			if (verbose)
				error(0, "adding key=%s data=%s\n", iline, p);
		}
		k->data = iline;
		k->size = strlen(iline);
		if (tacspec) {
			d->data = &element;
			d->size = sizeof(element);
		} else {
			d->data = p;
			d->size = strlen(p);
		}
		if ((*db->put) (db, k, d, NULL))
			error(0, "can't put key %s to db %s (input line %d)\n",
				iline, dbname, line);
	}
	(*db->close) (db);
}

/*
 * Edit entry
 */
void    edit(user)
	char    *user;
{
	DBT     k[1], d[1];
	char    *temp, *editor, buf[BUFSIZ];
	char    proto[] = "/tmp/tdbXXXXXX";
	FILE    *fp;
	int	i;
	struct in_addr iptmp;

	k->size = strlen(user);
	k->data = user;

	if ((*db->get) (db, k, d, NULL))
		error(1, "can't get db entry for user %s\n", user);
	temp = (char *) mktemp(proto);
	if (!(fp = fopen(temp, "w")))
err:
		error(1, "can't create/reopen temp file %s\n", temp);
	(void) fwrite(k->data, 1, k->size, fp);
	if (tacspec) {
		fputc(' ',fp);
		for (i=k->size; i<10; i++) fputc(' ',fp);
		iptmp.s_addr=((struct udb_data *) d->data)->addr;
		fprintf(fp, "time=%i\taddr=%s\n", (int)((struct udb_data *) d->data)->time, inet_ntoa(iptmp));
	} else {
		(void) putc('\t', fp);
		(void) fwrite(d->data, 1, d->size, fp);
		(void) putc('\n', fp);
	}
	(void) fclose(fp);
	editor = (char *) getenv("EDITOR");
	if (!editor)
		editor = "vi";
	(void) sprintf(buf, "%s %s", editor, temp);
	if(system(buf))
		error(1, "error while executing editor\n");
	(void) close(0);
	if (!fopen(temp, "r"))
		goto err;
	(void) unlink(temp);
	/* gluk ... */
	(void) sprintf(buf, "%s.bak", temp);
	(void) unlink(buf);
	(void) dbappend();
}

/*
 * Change password for user
 */
void    change_pass(docrypt, user, pass)
	int     docrypt;
	char    *user, *pass;
{
	DBT     k[1], d[1];
	char    buf[BUFSIZ], *p, salt[2], *np1, np2[BUFSIZ];
	int     ntries, offs;

	k->size = strlen(user);
	k->data = user;

	if ((*db->get) (db, k, d, NULL))
		error(1, "can't get db entry for user %s\n", user);

	for (p = d->data, offs = 0; offs < d->size; offs++, p++)
		if (*p == '|')
			break;
	salt[0] = 'A' + (getpid() % 17);
	salt[1] = 'a' + (time(NULL) % 17);
	ntries = 0;
	if (!pass) {
		for (;;) {
			np1 = getpass("Enter new password: ");
			(void) strcpy(np2, np1);
			np1 = getpass("Repeat new password: ");
			if (!strcmp(np1, np2))
				break;
			error(0, "Passwords mismatch; please try again.\n");
			if (ntries++ > 2)
				error(1, "\nAborted.\n");
		}
		np1 = crypt(np2, salt);
	} else
		np1 = docrypt ? crypt(pass, salt) : pass;
	(void) strcpy(buf, np1);
	if (*p == '|')
		(void) strncat(buf, p, d->size - offs);
	d->data = buf;
	d->size = strlen(buf);

	if ((*db->put) (db, k, d, NULL))
		error(1, "can't update key %s\n", user);

	(*db->close)(db);
}

void error(lvl, s, a1, a2, a3, a4)
	char    *s;
	int     lvl, a1, a2, a3, a4;
{
	fprintf(stderr, s, a1, a2, a3, a4);
	if (lvl)
		exit(lvl);
}
